//********************************************************
//
// Text Reporting Component 2.0. Simple Working Library
// Alex Shepetko (c) 1995-97.
//
// E-mail: alex_shepetko@hotmail.com
//
//********************************************************

unit Nsrep;

interface

uses
  SysUtils, WinProcs, WinTypes, Controls, Classes, Forms, DB,
  Dialogs, StdCtrls;

type
 JustifyType = ( JLeft, JMiddle, JRight );
 TCharSet = Set of Char;
 TSReportEvent = procedure of object;
 TPrintTo = (toFile, toDevice);

 TCustomSubDetail = class (TComponent)
 protected
  FSource: TDataSource;
  FBody, FCalc: TStringList;
  FCount: Integer;
  FChange: TSReportEvent;
  FEndHgt, FStartHgt: Integer;
 private
  procedure SetBody (Value: TStringList);
  procedure SetCalc (Value: TStringList);
  procedure SetSource (Value: TDataSource);
 public
  MaxMemo, MemoIndex: Integer;
  DetailMemos: String;
  StartPage, EndPage: TStringList;

  constructor Create (AOwner: TComponent); override;
  destructor Destroy;
  procedure AddCalcs;
  procedure InitCalcs;
  procedure ChangeData;
  procedure GetMaxMemoLines;

  property Body: TStringList read FBody write SetBody;
  property CalcColumns: TStringList read FCalc write SetCalc;
  property DataSource: TDataSource read FSource write SetSource;
  property DetCount: Integer read FCount write FCount;
  property EndHeight: Integer read FEndHgt write FEndHgt;
  property StartHeight: Integer read FStartHgt write FStartHgt;

  property OnChangingData: TSReportEvent read FChange write FChange;
 end;

 TNewCustomSReport = class (TComponent)
 protected
  {Variables}
  FForm, FOut: String;
  FSource: TDataSource;
  StartPage, EndPage, RepFile, FCalc, FAnswer: TStringList;
  FPrintTo: TPrintTo;
  FStartHgt, FEndHgt, PrnLns, MaxMemo: Integer;
  MemoIndex, FCount, FPageHgt: Integer;
  FSub: TCustomSubDetail;
  FAfterDet, FBeforeDetail, FPagePrinted: TSReportEvent;
  WasMemo, SubOrNot: Boolean;
 private
  {Private variables}
  DetailMemos: String;

  {Additional methods}
  procedure SetSource (Value: TDataSource);
  procedure SetAnswer (Value: TStringList);
  procedure SetCalc (Value: TStringList);
  procedure SetSub (Value: TCustomSubDetail);
 public
  {Methods}
  constructor Create (AOwner: TComponent); override;
  destructor Destroy;
  procedure PrintRep;
  procedure PrintOneStr (St: String);
  function PrintDet: Integer;
  procedure GetMaxMemoLines;
  function  GetLineOfMemo (FName: String; LIndex: Integer): String;
  function TranslateVar (VarText: String): String;
  function SAnsiToOEM (St: String): String;
  function PrintSubDet (St: String): Integer;
  procedure PrintSubBand;

  {Properties}
  property Answers: TStringList read FAnswer write SetAnswer;
  property DataSource: TDataSource read FSource write SetSource;
  property EndHgt: Integer read FEndHgt write FEndHgt;
  property FormName: String read FForm write FForm;
  property OutFile: String read FOut write FOut;
  property PageHeight: Integer read FPageHgt write FPageHgt;
  property PrintTo: TPrintTo read FPrintTo write FPrintTo;
  property StartHgt: Integer read FStartHgt write FStartHgt;
  property SubDetail: TCustomSubDetail read FSub write SetSub;

  {Calculated properties}
  property CalcColumns: TStringList read FCalc write SetCalc;
  property DetCount: Integer read FCount write FCount;

  {Events}
  property AfterDetails: TSReportEvent read FAfterDet
                             write FAfterDet;
  property BeforeDetail: TSReportEvent read FBeforeDetail
                             write FBeforeDetail;
  property PagePrinted: TSReportEvent read FPagePrinted
                             write FPagePrinted;
 end;

 TNewSReport = class (TNewCustomSReport)
 published
  property Answers;
  property DataSource;
  property EndHgt;
  property FormName;
  property OutFile;
  property PageHeight;
  property PrintTo;
  property StartHgt;
  property SubDetail;

  property AfterDetails;
  property BeforeDetail;
  property PagePrinted;
 end;

 TNewSRSubDetail = class (TCustomSubDetail)
 published
  property Body;
  property DataSource;
  property EndHeight;
  property StartHeight;
  property OnChangingData;
 end;

{Additional variables}
var
 Out: TextFile;
 Is_EndPage: Boolean;

function Justify( St: String; Len: Byte; Just: JustifyType ): String;

{Register components procedure}
procedure Register;

implementation

procedure Register;
begin
 RegisterComponents ('Standard', [TNewSReport, TNewSRSubDetail]);
end;

procedure SetBinaryMode;
begin
 {Set Out-file to binary mode}
 asm
  MOV   BX, offset(Out)
  MOV   AX,4400H
  INT   21H
  OR    DL,20H
  MOV   DH,0
  MOV   AX,4401H
  INT   21H
 end;
end;

{TCustomSReprot ----------------------------------------------}

constructor TNewCustomSReport.Create (AOwner: TComponent);
begin
 inherited Create (AOwner);

 {Intializing defaults}

 SubOrNot := False;
 WasMemo := False;
 FormName := '';
 OutFile := '';
 FAnswer := TStringList.Create;
 FCalc := TStringList.Create;
 RepFile := TStringList.Create;
 StartPage := TStringList.Create;
 EndPage := TStringList.Create;
 FStartHgt := 1;
 FEndHgt := 1;
 PrintTo := toDevice;
 PageHeight := 50;
end;

destructor TNewCustomSReport.Destroy;
begin
 {Disposing created objects}

 FAnswer.Free;
 FCalc.Free;
 RepFile.Free;
 StartPage.Free;
 EndPage.Free;
 FSource := nil;
 FSub := nil;

 inherited Destroy;
end;

procedure TNewCustomSReport.PrintRep;
var
 DetailNum, a: Integer;
 DetailSt: String;
begin
 {Name of form is absent}

 if FForm = '' then
 begin
  MessageDlg ('Name of form is absent', mtError, [mbOk], 0);
  Exit;
 end;

 {Output filename is absent when print to file is set}

 if (FPrintTo = toFile) and (FOut = '') then
 begin
  MessageDlg ('Output filename is absent when print to file is set',
                                mtError, [mbOk], 0);
  Exit;
 end;

 {The height of the page is zero}

 if PageHeight = 0 then
 begin
  MessageDlg ('The height of the page is zero', mtError, [mbOk], 0);
  Exit;
 end;

 {Height of start of page is more than height of
  the page}

 if StartHgt > PageHeight then
 begin
  MessageDlg ('Bad start height', mtError, [mbOk], 0);
  Exit;
 end;

 {The file of form not found}

 if FileSearch (FForm, '') = '' then
 begin
  MessageDlg ('Form file is absent', mtError, [mbOk], 0);
  Exit;
 end;

 if FPrintTo = toFile then
 begin
  AssignFile (Out, FOut);
 end
 else
 begin
  AssignFile (Out, 'LPT1');
  SetBinaryMode;
 end;

 RepFile.LoadFromFile (FForm);
 ReWrite (Out);

 {Which string you want to repeat}

 DetailNum := -1;

 DetailMemos := '';

 {Reading a start of page}

 StartPage.Clear;

 For a := 0 to FStartHgt - 1 do
 begin
  StartPage.Add (RepFile[0]);
 end;

 {Reading an end of page}

 EndPage.Clear;

 For a := 0 to FEndHgt - 1 do
 begin
  EndPage.Add (RepFile[0]);
  RepFile.Delete (0);
 end;

 if FSub <> nil then
 begin
  FSub.Body.Clear;

  DetailNum := -1;

  For a := 0 to RepFile.Count - 1 do
  begin
   if (RepFile[a][1] = '>') and not (RepFile[a] = '') then
   begin
    if DetailNum = -1 then
        DetailNum := a;

    FSub.Body.Add (Copy (RepFile [a], 2, Length (RepFile [a]) - 1));
   end;
  end;

  a := 0;

  While a <= RepFile.Count - 1 do
  begin
   if (RepFile[a][1] = '>') and not (RepFile[a] = '') then
    RepFile.Delete (a)
   else
    Inc (a);
  end;
 end;

 PrnLns := 0;

 {Printing before detail string}

 For a := 0 to DetailNum - 1 do
 begin
  PrintOneStr( RepFile[a] );
 end;

 {Printing detail string}

 FCount := 0;
 FCalc.Clear;

 if (FSource <> nil) and (FSource.DataSet <> nil) and
  FSource.DataSet.Active and (DetailNum <> -1) and (FSub <> nil) then
 begin

  For a := 0 to FSource.DataSet.FieldCount - 1 do
  begin
   FCalc.Add ('0');
  end;

  FSource.DataSet.First;

  While not FSource.DataSet.EOF do
  begin
   GetMaxMemoLines;

   PrintDet;

   FSource.DataSet.Next;
  end;

  if Assigned (FAfterDet) then
    FAfterDet;
 end;

 {Printing after detail string}

 For a := DetailNum to RepFile.Count - 1 do
 begin
  PrintOneStr( RepFile[a] );
 end;

 CloseFile (Out);
end;

procedure TNewCustomSReport.GetMaxMemoLines;
var
 SList: TStringList;
 a: Integer;
begin
 MaxMemo := 0;

 SList := TStringList.Create;

 For a := 0 to FSource.DataSet.FieldCount - 1 do
 begin
  if FSource.DataSet.Fields[a].DataType = ftMemo then
  begin
   SList.Assign (FSource.DataSet.Fields[a]);
   if (SList.Count > MaxMemo) and (Pos ('{@' +
     FSource.DataSet.Fields[a].FieldName, DetailMemos) <> 0) then
       MaxMemo := SList.Count;
  end;
 end;

 SList.Free;
end;

function TNewCustomSReport.GetLineOfMemo (FName: String;
                              LIndex: Integer): String;
var
 SList: TStringList;
begin
 Result := '';

 SList := TStringList.Create;

 if (FSource.DataSet.FieldByName (FName).DataType = ftMemo) and
            SubOrNot then
 begin
  SList.Assign (FSource.DataSet.FieldByName (FName));

  if SList.Count > LIndex then
    Result := SList[LIndex];
 end;

 if (FSource.DataSet.FieldByName (FName).DataType = ftMemo) and
       not SubOrNot then
 begin
  SList.Assign (FSub.DataSource.DataSet.FieldByName (FName));

  if SList.Count > LIndex then
    Result := SList[LIndex];

  WasMemo := True;
 end;

 SList.Free;
end;

procedure TNewCustomSReport.SetSource (Value: TDataSource);
begin
 FSource := Value;
end;

procedure TNewCustomSReport.SetSub (Value: TCustomSubDetail);
begin
 FSub := Value;
end;

procedure TNewCustomSReport.SetAnswer (Value: TStringList);
begin
 FAnswer.Assign (Value);
end;

procedure TNewCustomSReport.SetCalc (Value:TStringList);
begin
 FCalc.Assign (Value);
end;

procedure TNewCustomSReport.PrintOneStr (St: String);
var
 S, V: String;
 a: Integer;
 Is_V: Boolean;
begin
 if St = DetailMemos then
    Exit;

 S := '';

 Is_v := False;

 For a := 1 to Length (St) do
 begin
  if St[a] = '{' then
  begin
   V := St[a];
   Is_V := True;
   Continue;
  end;

  if Is_V then
  begin
   V := V + St[a];
  end;

  if St[a] = '}' then
  begin
   S := S + TranslateVar (V);
   Is_V := False;
   V := '';
   Continue;
  end;

  if not Is_v then
  begin
   S := S + St[a];
  end;
 end;

 WriteLn (Out, S);

 if Is_EndPage then Exit;

 Inc (PrnLns);

 if (PrnLns = FPageHgt) then
 begin
   Is_EndPage := True;

  if SubOrNot then
  begin
   For a := 0 to FSub.EndHeight - 1 do
   begin
    PrintOneStr (FSub.EndPage[a]);
   end;

   WriteLn (Out, #$0C);

   If Assigned (FPagePrinted) then
      FPagePrinted;
   PrnLns := 0;

   Is_EndPage := False;

   For a := 0 to FSub.StartHeight - 1 do
   begin
    PrintOneStr (FSub.StartPage[a]);
   end;
  end
  else
  begin
   For a := 0 to FEndHgt - 1 do
   begin
    PrintOneStr (EndPage[a]);
   end;

   WriteLn (Out, #$0C);

   If Assigned (FPagePrinted) then
      FPagePrinted;
   PrnLns := 0;

   Is_EndPage := False;

   For a := 0 to FStartHgt - 1 do
   begin
    PrintOneStr (StartPage[a]);
   end;
  end;
 end;
end;

function TNewCustomSReport.PrintSubDet (St: String): Integer;
var
 a: Integer;
 S: String;
begin
 if Assigned (FBeforeDetail) then
      FBeforeDetail;

 St := Copy (St, 2, Length (St) - 1);

 FSub.MemoIndex := 0;

 SubOrNot := True;
 PrintOneStr (St);

 if FSub.MemoIndex < FSub.MaxMemo - 1 then
 begin
  repeat
   Inc (FSub.MemoIndex);

   PrintOneStr (Copy (FSub.DetailMemos, 2, Length (FSub.DetailMemos) - 1));
  until FSub.MemoIndex = FSub.MaxMemo - 1;
 end;
 SubOrNot := False;

 Result := FSub.MemoIndex + 1;
end;

procedure TNewCustomSReport.PrintSubBand;
var
 DetailNum, a: Integer;
 DetailSt: String;
begin
 {Which string you want to repeat}

 DetailNum := FSub.Body.Count;

 FSub.DetailMemos := '';

 {Reading a start of page}

 FSub.StartPage := TStringList.Create;

 For a := 0 to FStartHgt - 1 do
 begin
  FSub.StartPage.Add (FSub.Body[a]);
 end;

 {Reading an end of page}

 FSub.EndPage := TStringList.Create;

 For a := FSub.StartHeight to FSub.StartHeight +
              FSub.EndHeight - 1 do
 begin
  FSub.EndPage.Add (FSub.Body[a]);
 end;

 For a := FSub.StartHeight + FSub.EndHeight to FSub.Body.Count - 1 do
 begin
  if (FSub.Body[a][1] = '>') and not (FSub.Body[a] = '') then
    DetailNum := a;

  if (FSub.Body[a][1] = '~') and not (FSub.Body[a] = '') then
    FSub.DetailMemos := FSub.Body[a];
 end;

 PrnLns := 0;

 {Printing before detail string}

 For a := FSub.StartHeight + FSub.EndHeight to DetailNum - 1 do
 begin
  WasMemo := False;

  PrintOneStr( FSub.Body[a] );

   if WasMemo then
       Inc (MemoIndex);
 end;

 {Printing detail string}

 FSub.InitCalcs;

 if (FSub.DataSource <> nil) and (FSub.DataSource.DataSet <> nil) and
  FSub.DataSource.DataSet.Active and (DetailNum < FSub.Body.Count) then
 begin
  DetailSt := FSub.Body[DetailNum];
  FSub.DataSource.DataSet.First;

  While not FSub.DataSource.DataSet.EOF do
  begin
   FSub.GetMaxMemoLines;

   PrintSubDet (DetailSt);

   FSub.AddCalcs;
   FSub.DataSource.DataSet.Next;
  end;
 end;

 {Printing after detail string}

 if DetailNum <> FSub.Body.Count then
 begin
  For a := DetailNum + 1 to FSub.Body.Count - 1 do
  begin
   WasMemo := False;

   PrintOneStr (FSub.Body[a]);

   if WasMemo then
       Inc (MemoIndex);
  end;
 end;

 FSub.ChangeData;

 FSub.StartPage.Free;
 FSub.EndPage.Free;
end;

function TNewCustomSReport.PrintDet: Integer;
var
 a: Integer;
 S, St: String;
begin
 if Assigned (FBeforeDetail) then
      FBeforeDetail;

 FSub.InitCalcs;

 PrintSubBand;

 St := Copy (St, 2, Length (St) - 1);

 Inc (FCount);

 For a := 0 to FSource.DataSet.FieldCount - 1 do
 begin
  if FSource.DataSet.Fields[a].DataType in
       [ftSmallint, ftInteger, ftWord, ftFloat, ftCurrency] then
  begin
   S := FCalc[a];
   S := FloatToStrF (StrToFloat (S) +
           FSource.DataSet.Fields[a].AsFloat, ffFixed, 12, 10);

   While (S[Length (S)] = '0')  and (Length (S) > 0) do
     S := Copy (S, 1, Length (S) - 1);

   if S = '' then
     S := '0';

   While (S[Length (S)] = '.') and (Length (S) > 0) do
     S := Copy (S, 1, Length (S) - 1);

   if S = '' then
     S := '0';

   FCalc[a] := S;
  end;
 end;
end;

function Justify( St: String; Len: Byte; Just: JustifyType ): String;
var
 J: JustifyType;
begin
 Result := St;
 if Just = JLeft then
 begin
  while Length( Result ) < Len do
    Result := Result + ' ';
 end;

 if Just = JRight then
 begin
  while Length( Result ) < Len do
    Result := ' ' + Result;
 end;

 J := JLeft;
 if Just = JMiddle then
 begin
  while Length( Result ) < Len do
  begin
   Result := Justify( Result, Length( Result ) + 1, J );
   if J = JLeft then
     J := JRight
   else
     J := JLeft;
  end;
 end;

 if Length( St ) > Len then
     result := copy( St, 1, len );
end;

function TNewCustomSReport.TranslateVar (VarText: String): String;

procedure CutSt( St, Divider: String; var ListSt: TStringList );
var
 S: String;
 Pos: Integer;
begin
 Pos := 1;
 S := '';

 ListSt.Clear;

 While Pos <= Length( St ) do
 begin
  if St[ Pos ] = Divider then
  begin
   ListSt.Add( S );
   S := '';
  end
  else
   S := S + St[ Pos ];

  Inc( Pos );
 end;

 ListSt.Add( S );
end;


function DelOnlyNot( S: String; ChSet: TCharSet ): String;
var
 Res: String;
 a: Integer;
begin
 Res := '';
 For a := 1 to Length( S ) do
   if S[ a ] in ChSet then
       Res := Res + S[ a ];
 DelOnlyNot := Res;
end;

const
 nName = 0;
 nJust = 1;
 nLen  = 2;

 sLft: TCharSet = ['L', 'l'];
 sRgt: TCharSet = ['R', 'r'];
 sMid: TCharSet = ['M', 'm'];

var
 SList: TStringList;
 J: JustifyType;
 Len: Integer;
 S, Value: String;
begin
 Len := 0;

 SList := TStringList.Create;

 VarText := Copy (VarText, 2, Length (VarText) - 2);

 CutSt (VarText, ',', SList);

 J := jLeft;

 if SList[nJust][1] in sRgt then
       J := jRight;
 if SList[nJust][1] in sMid then
       J := jMiddle;

 S := DelOnlyNot (SList[nLen], ['0'..'9']);

 if Length (S) = 0 then
   S := '0';
 Len := StrToInt (S);

 S := SList[nName];

 if S[1] = '@' then
 begin
  S := Copy (S, 2, Length (S) - 1);

  if SubOrNot then
  begin
   if FSub.DataSource.DataSet.FieldByName (S).DataType = ftMemo then
   begin
    Value := GetLineOfMemo (S, FSub.MemoIndex);
   end
   else
   begin
    Value := FSub.DataSource.DataSet.FieldByName (S).Text;
   end;
  end
  else
  begin
   if FSource.DataSet.FieldByName (S).DataType = ftMemo then
   begin
    Value := GetLineOfMemo (S, MemoIndex);
   end
   else
   begin
    Value := FSource.DataSet.FieldByName (S).Text;
   end;
  end;
 end
 else
 begin
  S := DelOnlyNot (S, ['0'..'9']);
  if Length (S) = 0 then
    S := '0';

  Value := FAnswer[StrToInt (S)];
 end;

 Value := SAnsiToOEM (Value);

 if Len <> 0 then
   Value := Justify (Value, Len, J);

 SList.Free;

 Result := Value;
end;

function TNewCustomSReport.SAnsiToOEM (St: String): String;
var
 Decoding, Res: Array[ 0..255 ] of Char;
 a, Len: Integer;
begin
 For a := 1 to Length( St ) do
 begin
  Decoding[a - 1] := St[a];
  Decoding[a] := #0;
 end;

 AnsiToOEM (Decoding, Res);

 Result := '';

 For a := 0 to Length (St) - 1 do
 begin
  Result := Result + Res[a];
 end;
end;

{TCustomSubDetail --------------------------------------------}

constructor TCustomSubDetail.Create (AOwner: TComponent);
begin
 inherited Create (AOwner);

 StartHeight := 1;
 EndHeight := 1;
 FBody := TStringList.Create;
 FCalc := TStringList.Create;
end;

destructor TCustomSubDetail.Destroy;
begin
 FBody.Free;
 FCalc.Free;
 FSource := nil;

 inherited Destroy;
end;

procedure TCustomSubDetail.InitCalcs;
var
 a: Integer;
begin
 FCalc.Clear;

 For a := 1 to DataSource.DataSet.FieldCount do
 begin
  FCalc.Add ('0');
 end;

 FCount := 0;
end;

procedure TCustomSubDetail.AddCalcs;
var
 a: Integer;
begin
 For a := 1 to DataSource.DataSet.FieldCount do
 begin
  if not (DataSource.DataSet.Fields [a - 1].DataType in
   [ftSmallint, ftInteger, ftWord, ftFloat, ftCurrency]) then
    Continue;

  FCalc [a - 1] := FloatToStr (StrToFloat (FCalc [a - 1]) +
                  DataSource.DataSet.Fields [a - 1].AsFloat);
 end;

 Inc (FCount);
end;

procedure TCustomSubDetail.SetBody (Value: TStringList);
begin
 FBody.Assign (Value);
end;

procedure TCustomSubDetail.SetCalc (Value: TStringList);
begin
 FCalc.Assign (Value);
end;

procedure TCustomSubDetail.SetSource (Value: TDataSource);
begin
 FSource := Value;
end;

procedure TCustomSubDetail.GetMaxMemoLines;
var
 SList: TStringList;
 a: Integer;
begin
 MaxMemo := 0;

 SList := TStringList.Create;

 For a := 0 to FSource.DataSet.FieldCount - 1 do
 begin
  if FSource.DataSet.Fields[a].DataType = ftMemo then
  begin
   SList.Assign (FSource.DataSet.Fields[a]);
   if (SList.Count > MaxMemo) and (Pos ('{@' +
     FSource.DataSet.Fields[a].FieldName, DetailMemos) <> 0) then
       MaxMemo := SList.Count;
  end;
 end;

 SList.Free;
end;

procedure TCustomSubDetail.ChangeData;
begin
 if Assigned (FChange) then FChange;
end;

end.
